#!/usr/bin/env python
# Copyright (c) 2002 Joao Prado Maia. See the LICENSE file for more information.
# Substantial changes in 2003 by Mikael "MC" Cardell <mc@hack.org>.
# $Id: $
import MySQLdb
import time
import re
import settings
import strutil
import mime

# we don't need to compile the regexps everytime..
singleline_regexp = re.compile("^\.", re.M)
from_regexp = re.compile("^From:(.*)", re.M)
subject_regexp = re.compile("^Subject:(.*)", re.M)
references_regexp = re.compile("^References:(.*)<(.*)>", re.M)

class Papercut_Storage:
    """
    Storage Backend interface for saving the article information in a
    MySQL database mapped to the tables used by Gnuheter
    (http://gnuheter.com/), a hacked version of PHPNuke.
    """

    def __init__(self):
        self.conn = MySQLdb.connect(host=settings.dbhost, db=settings.dbname, user=settings.dbuser, passwd=settings.dbpass)
        self.cursor = self.conn.cursor()

    def quote_string(self, text):
        """Quotes strings the MySQL way."""
        return text.replace("'", "\\'")

    def get_body(self, lines):
        pass

    def get_header(self, lines):
        pass

    def get_message_id(self, msg_num, group):
        return '<%s@%s>' % (msg_num, group)

    def group_exists(self, group_name):
	# Massage group_name if ending with .d
	if group_name[-2:] == ".d":
	    group_name = group_name[:-2]
        stmt = """
                SELECT
                    COUNT(*) AS check
                FROM
                    topics
                WHERE
                    topicname='%s'""" % (group_name)
        self.cursor.execute(stmt)
        return self.cursor.fetchone()[0]

    def article_exists(self, group_name, style, range):
        # table_name = self.get_table_name(group_name)
        stmt = """
                SELECT
                    COUNT(*) AS check
                FROM
                    stories
                WHERE
                    """
        if style == 'range':
            stmt = "%s id > %s" % (stmt, range[0])
            if len(range) == 2:
                stmt = "%s AND id < %s" % (stmt, range[1])
        else:
            stmt = "%s id = %s" % (stmt, range[0])
        self.cursor.execute(stmt)
        return self.cursor.fetchone()[0]

    def get_group_stats(self, group_name):
	if group_name[-2:] == ".d":
	    stmt = """
	        SELECT
                    COUNT(tid) AS total,
                    IF(MAX(tid) IS NULL, 0, MAX(tid)) AS maximum,
                    IF(MIN(tid) IS NULL, 0, MIN(tid)) AS minimum
                FROM
                    comments, topics
                WHERE
		  topics.topicname='%s'""" % (group_name[:-2])
	else:
	    stmt = """
	        SELECT
		  COUNT(sid) AS total,
		  IF(MAX(sid) IS NULL, 0, MAX(sid)) AS maximum,
		  IF(MIN(sid) IS NULL, 0, MIN(sid)) AS minimum
	        FROM
		  stories, topics
	        WHERE
		  topicname='%s'""" % (group_name)
	
	num_rows = self.cursor.execute(stmt)
        return self.cursor.fetchone()

    def get_GROUP(self, group_name):
	#table_name = self.get_table_name(group_name)

	result = self.get_group_stats(group_name)

	# result = self.get_group_stats("stories")

	return (result[0], result[2], result[1])

    def get_LIST(self):
        # First find all the moderated groups.
        stmt = """
                SELECT
                    topicname, count(sid)
                FROM
                    stories, topics
                WHERE
                    topics.topicid=stories.topic
                GROUP BY
                    topicname
                ORDER BY
                    topicname ASC"""
        self.cursor.execute(stmt)
        result = list(self.cursor.fetchall())
        if len(result) == 0:
            return ""
        else:
            lists = []
            for topicname, max in result:
                 if settings.server_type == 'read-only':
                     lists.append("%s %s %s n" % (topicname, max, 0))
                 else:
                     lists.append("%s %s %s y" % (topicname, max, 0))
            
            # Now find the corresponding discussion groups where no
            # original posts can be done.

            stmt = """
                SELECT
                    topicname, count(tid)
                FROM
                    comments, stories, topics
                WHERE
                    topics.topicid=stories.topic and comments.sid=stories.sid
                GROUP BY
                    topicname
                ORDER BY
                    topicname ASC"""
            self.cursor.execute(stmt)
            result = list(self.cursor.fetchall())
            if len(result) != 0:
                for topicname, max in result:
                    if settings.server_type == 'read-only':
                        lists.append("%s.d %s %s n" % (topicname, max, 0))
                    else:
                        lists.append("%s.d %s %s y" % (topicname, max, 0))
            return "\r\n".join(lists)

    def get_ARTICLE(self, group_name, id):

        headers = []

        # Check if the request article is a story or a comment. If the
        # group name ends with ".d" it's a comment.

	if group_name[-2:] == ".d":
            stmt = """
                   SELECT
                       tid,
                       pid,
                       sid,
                       date,
                       name,
                       email,
                       url,
                       host_name,
                       subject,
                       comment,
                       score,
                       reason
                    FROM
                       comments
                    WHERE
                       tid=%s""" % (id)
        else:
            stmt = """
                   SELECT
                       sid,
                       aid,
                       title,
                       time,
                       hometext,
                       bodytext
                    FROM
                       stories
                    WHERE
                       sid=%s""" % (id)

        num_rows = self.cursor.execute(stmt)
        if num_rows == 0:
            return None
        result = list(self.cursor.fetchone())

	if group_name[-2:] == ".d":
	    headers.append("Path: %s" % (settings.nntp_hostname))
	    headers.append("From: %s <%s>" % (result[4], result[5]))
	    headers.append("Newsgroups: %s" % (group_name))
	    headers.append("Date: %s" % (result[3]))
	    headers.append("Subject: %s" % (result[8]))
	    return ("\r\n".join(headers), strutil.format_body(result[9]))
	else:
	    headers.append("Path: %s" % (settings.nntp_hostname))
	    headers.append("From: %s" % (result[1]))
	    headers.append("Newsgroups: %s" % (group_name))
	    headers.append("Date: %s" % (result[3]))
	    headers.append("Subject: %s" % (result[2]))
	    headers.append("Message-ID: <%s@%s>" % (result[0], group_name))
	    headers.append("Xref: %s %s:%s" % (settings.nntp_hostname, group_name, result[0]))

	    return ("\r\n".join(headers), strutil.format_body(result[3] + result[4]))

    def get_HEAD(self, group_name, id):
        #table_name = self.get_table_name(group_name)
        stmt = """
                SELECT
                    sid,
                    aid,
                    title,
                    time
                FROM
                    stories
                WHERE
                    sid=%s""" % (id)
        num_rows = self.cursor.execute(stmt)
        if num_rows == 0:
            return None
        result = list(self.cursor.fetchone())
        headers = []
        headers.append("Path: %s" % (settings.nntp_hostname))
        headers.append("From: %s" % (result[1]))
        headers.append("Newsgroups: %s" % (group_name))
        headers.append("Date: %s" % (result[3]))
        headers.append("Subject: %s" % (result[2]))
        headers.append("Message-ID: <%s@%s>" % (result[0], group_name))

        # MC:
        # I want this to be something more like this:
        # headers.append("Message-ID: <%s.%s@%s>" % (result[0], group_name, settings.nntp_hostname))
        # in case we distribute it, but that breaks the current HEAD and ARTICLE implementation.

        headers.append("Xref: %s %s:%s" % (settings.nntp_hostname, group_name, result[0]))

#        if result[4] != 0:
#            headers.append("References: <%s@%s>" % (result[4], group_name))
        return "\r\n".join(headers)


#     def get_BODY(self, group_name, id):
#         table_name = self.get_table_name(group_name)
#         stmt = """
#                 SELECT
#                     body
#                 FROM
#                     %s
#                 WHERE
#                     id=%s""" % (table_name, id)
#         num_rows = self.cursor.execute(stmt)
#         if num_rows == 0:
#             return None
#         else:
#             return strutil.format_body(self.cursor.fetchone()[0])

    def get_XOVER(self, group_name, start_id, end_id='ggg'):
	if group_name[-2:] == ".d":
	    pass
	else:
	    stmt = """
                SELECT
                    sid,
                    title,
                    aid,
                    time,
                    hometext,
                    bodytext
                FROM
                    stories, topics
                WHERE
                    sid >= %s and stories.topic=topics.topicid
                    AND topics.topicname='%s'""" % (start_id, group_name)
	if end_id != 'ggg':
	    stmt = "%s AND sid <= %s" % (stmt, end_id)

        self.cursor.execute(stmt)
        result = list(self.cursor.fetchall())
        overviews = []
        for row in result:
            message_id = "<%s@%s>" % (row[0], group_name)
            xref = 'Xref: %s %s:%s' % (settings.nntp_hostname, group_name, \
				       row[0])

            reference = ""

            overviews.append("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (row[0], row[3], row[2], row[4], message_id, reference, "1", "1", xref))
        return "\r\n".join(overviews)


    # POST is only for comments for now: the 'stories' groups are moderated.
    # XXX Should add support for diary entries. XXX

    def do_POST(self, group_name, body, ip_address, username=''):
        name = from_regexp.search(body, 0).groups()[0].strip()
        subject = subject_regexp.search(body, 0).groups()[0].strip()

        if body.find('References') != -1:
            references = references_regexp.search(body, 1).groups()

            # Now figure out if the first reference refers to a
            # comment or a story.

            # A story is known as <s@topic>.

            # A comment is known as <s.c@topic>. re.compile(".*\..*@")

            # if references 

            # pid, void = references[-1].strip().split('@')

        body = mime.get_body(body)
        stmt = """
                INSERT INTO
                    comments
                (
                    pid,
                    sid,
                    date,
                    name,
                    email,
                    url,
                    host_name,
                    subject,
                    comment
                ) VALUES (
                    %s,
                    %s,
                    NOW(),
                    '%s',
                    '%s',
                    '%s',
                    '%s',
                    '%s',
                    '%s'
                )
                """ % (pid, 2, "mc", "mc@hack.org", "http://hack.org/~mc", "zoot.hack.org", "subject", body)

                # MC: Note that story id (sid) is hardcoded above!

        if not self.cursor.execute(stmt):
            return None
        else:
            # Now update the corresponding row in 'stories' to add a comment.

            stmt = """
                UPDATE
                    stories
                SET
                    comments==comments+1
                WHERE
                    sid=2
                    """
        if not self.cursor.execute(stmt):
            return None
        else:
            return 1
